@@ -1,13 +1,20 @@
 use super::*;
+use core::convert::TryInto;
+
+#[cfg(feature = "serde")]
+use core::marker::PhantomData;
+#[cfg(feature = "serde")]
+use serde::de::{
+  Deserialize, Deserializer, Error as DeserializeError, SeqAccess, Visitor,
+};
+#[cfg(feature = "serde")]
+use serde::ser::{Serialize, SerializeSeq, Serializer};
 
 /// Helper to make an `ArrayVec`.
 ///
 /// You specify the backing array type, and optionally give all the elements you
 /// want to initially place into the array.
 ///
-/// As an unfortunate restriction, the backing array type must support `Default`
-/// for it to work with this macro.
-///
 /// ```rust
 /// use tinyvec::*;
 ///
@@ -54,8 +61,7 @@ macro_rules! array_vec {
 ///
 /// ## Construction
 ///
-/// If the backing array supports Default (length 32 or less), then you can use
-/// the `array_vec!` macro similarly to how you might use the `vec!` macro.
+/// You can use the `array_vec!` macro similarly to how you might use the `vec!` macro.
 /// Specify the array type, then optionally give all the initial values you want
 /// to have.
 /// ```rust
@@ -89,18 +95,27 @@ macro_rules! array_vec {
 /// assert_eq!(more_ints.len(), 2);
 /// ```
 #[repr(C)]
-#[derive(Clone, Copy, Default)]
+#[derive(Clone, Copy)]
 pub struct ArrayVec<A: Array> {
-  len: usize,
+  len: u16,
   pub(crate) data: A,
 }
 
+impl<A: Array> Default for ArrayVec<A> {
+  fn default() -> Self {
+    Self {
+      len: 0,
+      data: A::default(),
+    }
+  }
+}
+
 impl<A: Array> Deref for ArrayVec<A> {
   type Target = [A::Item];
   #[inline(always)]
   #[must_use]
   fn deref(&self) -> &Self::Target {
-    &self.data.as_slice()[..self.len]
+    &self.data.as_slice()[..self.len as usize]
   }
 }
 
@@ -108,7 +123,7 @@ impl<A: Array> DerefMut for ArrayVec<A> {
   #[inline(always)]
   #[must_use]
   fn deref_mut(&mut self) -> &mut Self::Target {
-    &mut self.data.as_slice_mut()[..self.len]
+    &mut self.data.as_slice_mut()[..self.len as usize]
   }
 }
 
@@ -129,6 +144,39 @@ impl<A: Array, I: SliceIndex<[A::Item]>> IndexMut<I> for ArrayVec<A> {
   }
 }
 
+#[cfg(feature = "serde")]
+#[cfg_attr(docs_rs, doc(cfg(feature = "serde")))]
+impl<A: Array> Serialize for ArrayVec<A>
+where
+  A::Item: Serialize,
+{
+  #[must_use]
+  fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+  where
+    S: Serializer,
+  {
+    let mut seq = serializer.serialize_seq(Some(self.len()))?;
+    for element in self.iter() {
+      seq.serialize_element(element)?;
+    }
+    seq.end()
+  }
+}
+
+#[cfg(feature = "serde")]
+#[cfg_attr(docs_rs, doc(cfg(feature = "serde")))]
+impl<'de, A: Array> Deserialize<'de> for ArrayVec<A>
+where
+  A::Item: Deserialize<'de>,
+{
+  fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+  where
+    D: Deserializer<'de>,
+  {
+    deserializer.deserialize_seq(ArrayVecVisitor(PhantomData))
+  }
+}
+
 impl<A: Array> ArrayVec<A> {
   /// Move all values from `other` into this vec.
   ///
@@ -284,7 +332,7 @@ impl<A: Array> ArrayVec<A> {
       return;
     }
 
-    let new_len = self.len + sli.len();
+    let new_len = self.len as usize + sli.len();
     assert!(
       new_len <= A::CAPACITY,
       "ArrayVec::extend_from_slice> total length {} exceeds capacity {}!",
@@ -292,7 +340,7 @@ impl<A: Array> ArrayVec<A> {
       A::CAPACITY
     );
 
-    let target = &mut self.data.as_slice_mut()[self.len..new_len];
+    let target = &mut self.data.as_slice_mut()[self.len as usize..new_len];
     target.clone_from_slice(sli);
     self.set_len(new_len);
   }
@@ -397,7 +445,7 @@ impl<A: Array> ArrayVec<A> {
   #[inline]
   pub fn try_insert(&mut self, index: usize, item: A::Item) -> Option<A::Item> {
     assert!(
-      index <= self.len,
+      index <= self.len as usize,
       "ArrayVec::try_insert> index {} is out of bounds {}",
       index,
       self.len
@@ -422,16 +470,13 @@ impl<A: Array> ArrayVec<A> {
   #[inline(always)]
   #[must_use]
   pub fn len(&self) -> usize {
-    self.len
+    self.len as usize
   }
 
   /// Makes a new, empty `ArrayVec`.
   #[inline(always)]
   #[must_use]
-  pub fn new() -> Self
-  where
-    A: Default,
-  {
+  pub fn new() -> Self {
     Self::default()
   }
 
@@ -452,7 +497,7 @@ impl<A: Array> ArrayVec<A> {
   pub fn pop(&mut self) -> Option<A::Item> {
     if self.len > 0 {
       self.len -= 1;
-      let out = take(&mut self.data.as_slice_mut()[self.len]);
+      let out = take(&mut self.data.as_slice_mut()[self.len as usize]);
       Some(out)
     } else {
       None
@@ -496,9 +541,9 @@ impl<A: Array> ArrayVec<A> {
   /// ```
   #[inline(always)]
   pub fn try_push(&mut self, val: A::Item) -> Option<A::Item> {
-    debug_assert!(self.len <= A::CAPACITY);
+    debug_assert!(self.len as usize <= A::CAPACITY);
 
-    let itemref = match self.data.as_slice_mut().get_mut(self.len) {
+    let itemref = match self.data.as_slice_mut().get_mut(self.len as usize) {
       None => return Some(val),
       Some(x) => x,
     };
@@ -585,7 +630,7 @@ impl<A: Array> ArrayVec<A> {
     new_len: usize,
     mut f: F,
   ) {
-    match new_len.checked_sub(self.len) {
+    match new_len.checked_sub(self.len as usize) {
       None => self.truncate(new_len),
       Some(new_elements) => {
         for _ in 0..new_elements {
@@ -624,12 +669,13 @@ impl<A: Array> ArrayVec<A> {
     }
 
     let mut rest = JoinOnDrop {
-      items: &mut self.data.as_slice_mut()[..self.len],
+      items: &mut self.data.as_slice_mut()[..self.len as usize],
       done_end: 0,
       tail_start: 0,
     };
 
-    for idx in 0..self.len {
+    let len = self.len as usize;
+    for idx in 0..len {
       // Loop start invariant: idx = rest.done_end + rest.tail_start
       if !acceptable(&rest.items[idx]) {
         let _ = take(&mut rest.items[idx]);
@@ -664,9 +710,12 @@ impl<A: Array> ArrayVec<A> {
         new_len,
         A::CAPACITY
       )
-    } else {
-      self.len = new_len;
     }
+
+    let new_len: u16 = new_len
+      .try_into()
+      .expect("ArrayVec::set_len> new length is not in range 0..=u16::MAX");
+    self.len = new_len;
   }
 
   /// Splits the collection at the point given.
@@ -687,12 +736,9 @@ impl<A: Array> ArrayVec<A> {
   /// assert_eq!(&av2[..], [2, 3]);
   /// ```
   #[inline]
-  pub fn split_off(&mut self, at: usize) -> Self
-  where
-    Self: Default,
-  {
+  pub fn split_off(&mut self, at: usize) -> Self {
     // FIXME: should this just use drain into the output?
-    if at > self.len {
+    if at > self.len() {
       panic!(
         "ArrayVec::split_off> at value {} exceeds length of {}",
         at, self.len
@@ -703,8 +749,10 @@ impl<A: Array> ArrayVec<A> {
     let split_len = moves.len();
     let targets = &mut new.data.as_slice_mut()[..split_len];
     moves.swap_with_slice(targets);
-    new.len = split_len;
-    self.len = at;
+
+    /* moves.len() <= u16::MAX, so these are surely in u16 range */
+    new.len = split_len as u16;
+    self.len = at as u16;
     new
   }
 
@@ -747,11 +795,11 @@ impl<A: Array> ArrayVec<A> {
     use core::ops::Bound;
     let start = match range.start_bound() {
       Bound::Included(x) => *x,
-      Bound::Excluded(x) => x + 1,
+      Bound::Excluded(x) => x.saturating_add(1),
       Bound::Unbounded => 0,
     };
     let end = match range.end_bound() {
-      Bound::Included(x) => x + 1,
+      Bound::Included(x) => x.saturating_add(1),
       Bound::Excluded(x) => *x,
       Bound::Unbounded => self.len(),
     };
@@ -795,12 +843,12 @@ impl<A: Array> ArrayVec<A> {
   #[inline]
   pub fn swap_remove(&mut self, index: usize) -> A::Item {
     assert!(
-      index < self.len,
+      index < self.len(),
       "ArrayVec::swap_remove> index {} is out of bounds {}",
       index,
       self.len
     );
-    if index == self.len - 1 {
+    if index == self.len() - 1 {
       self.pop().unwrap()
     } else {
       let i = self.pop().unwrap();
@@ -813,13 +861,20 @@ impl<A: Array> ArrayVec<A> {
   /// If the vec is already shorter than the input, nothing happens.
   #[inline]
   pub fn truncate(&mut self, new_len: usize) {
-    if needs_drop::<A::Item>() {
-      while self.len > new_len {
-        self.pop();
-      }
-    } else {
-      self.len = self.len.min(new_len);
+    if new_len >= self.len as usize {
+      return;
     }
+
+    if needs_drop::<A::Item>() {
+      let len = self.len as usize;
+      self.data.as_slice_mut()[new_len..len]
+        .iter_mut()
+        .map(take)
+        .for_each(drop);
+    }
+
+    /* new_len is less than self.len */
+    self.len = new_len as u16;
   }
 
   /// Wraps an array, using the given length as the starting length.
@@ -833,8 +888,9 @@ impl<A: Array> ArrayVec<A> {
   /// error, and you'll get the array back in the `Err`.
   #[inline]
   pub fn try_from_array_len(data: A, len: usize) -> Result<Self, A> {
+    /* Note(Soveu): Should we allow A::CAPACITY > u16::MAX for now? */
     if len <= A::CAPACITY {
-      Ok(Self { data, len })
+      Ok(Self { data, len: len as u16 })
     } else {
       Err(data)
     }
@@ -858,7 +914,7 @@ impl<A: Array> ArrayVec<A> {
   /// ```
   #[inline(always)]
   pub fn grab_spare_slice(&self) -> &[A::Item] {
-    &self.data.as_slice()[self.len..]
+    &self.data.as_slice()[self.len as usize..]
   }
 
   /// Obtain the mutable slice of the array _after_ the active memory.
@@ -874,7 +930,7 @@ impl<A: Array> ArrayVec<A> {
   /// ```
   #[inline(always)]
   pub fn grab_spare_slice_mut(&mut self) -> &mut [A::Item] {
-    &mut self.data.as_slice_mut()[self.len..]
+    &mut self.data.as_slice_mut()[self.len as usize..]
   }
 }
 
@@ -927,6 +983,7 @@ impl<'p, A: Array, I: Iterator<Item = A::Item>> Iterator
 {
   type Item = A::Item;
 
+  #[inline]
   fn next(&mut self) -> Option<A::Item> {
     if self.removal_start < self.removal_end {
       match self.replacement.next() {
@@ -979,6 +1036,7 @@ where
   A: Array,
   I: Iterator<Item = A::Item> + DoubleEndedIterator,
 {
+  #[inline]
   fn next_back(&mut self) -> Option<A::Item> {
     if self.removal_start < self.removal_end {
       match self.replacement.next_back() {
@@ -1066,11 +1124,16 @@ impl<A: Array> From<A> for ArrayVec<A> {
   /// If you want to select a length, use
   /// [`from_array_len`](ArrayVec::from_array_len)
   fn from(data: A) -> Self {
-    Self { len: data.as_slice().len(), data }
+    let len: u16 = data
+      .as_slice()
+      .len()
+      .try_into()
+      .expect("ArrayVec::from> lenght must be in range 0..=u16::MAX");
+    Self { len, data }
   }
 }
 
-impl<A: Array + Default> FromIterator<A::Item> for ArrayVec<A> {
+impl<A: Array> FromIterator<A::Item> for ArrayVec<A> {
   #[inline]
   #[must_use]
   fn from_iter<T: IntoIterator<Item = A::Item>>(iter: T) -> Self {
@@ -1084,8 +1147,8 @@ impl<A: Array + Default> FromIterator<A::Item> for ArrayVec<A> {
 
 /// Iterator for consuming an `ArrayVec` and returning owned elements.
 pub struct ArrayVecIterator<A: Array> {
-  base: usize,
-  len: usize,
+  base: u16,
+  tail: u16,
   data: A,
 }
 
@@ -1094,7 +1157,7 @@ impl<A: Array> ArrayVecIterator<A> {
   #[inline]
   #[must_use]
   pub fn as_slice(&self) -> &[A::Item] {
-    &self.data.as_slice()[self.base..self.len]
+    &self.data.as_slice()[self.base as usize..self.tail as usize]
   }
 }
 impl<A: Array> FusedIterator for ArrayVecIterator<A> {}
@@ -1102,7 +1165,8 @@ impl<A: Array> Iterator for ArrayVecIterator<A> {
   type Item = A::Item;
   #[inline]
   fn next(&mut self) -> Option<Self::Item> {
-    let slice = &mut self.data.as_slice_mut()[self.base..self.len];
+    let slice =
+      &mut self.data.as_slice_mut()[self.base as usize..self.tail as usize];
     let itemref = slice.first_mut()?;
     self.base += 1;
     return Some(take(itemref));
@@ -1110,12 +1174,13 @@ impl<A: Array> Iterator for ArrayVecIterator<A> {
   #[inline(always)]
   #[must_use]
   fn size_hint(&self) -> (usize, Option<usize>) {
-    let s = self.len - self.base;
+    let s = self.tail - self.base;
+    let s = s as usize;
     (s, Some(s))
   }
   #[inline(always)]
   fn count(self) -> usize {
-    self.len - self.base
+    self.size_hint().0
   }
   #[inline]
   fn last(mut self) -> Option<Self::Item> {
@@ -1123,14 +1188,16 @@ impl<A: Array> Iterator for ArrayVecIterator<A> {
   }
   #[inline]
   fn nth(&mut self, n: usize) -> Option<A::Item> {
-    let slice = &mut self.data.as_slice_mut()[self.base..self.len];
+    let slice = &mut self.data.as_slice_mut();
+    let slice = &mut slice[self.base as usize..self.tail as usize];
 
     if let Some(x) = slice.get_mut(n) {
-      self.base += n + 1;
+      /* n is in range [0 .. self.tail - self.base) so in u16 range */
+      self.base += n as u16 + 1;
       return Some(take(x));
     }
 
-    self.base = self.len;
+    self.base = self.tail;
     return None;
   }
 }
@@ -1138,20 +1205,29 @@ impl<A: Array> Iterator for ArrayVecIterator<A> {
 impl<A: Array> DoubleEndedIterator for ArrayVecIterator<A> {
   #[inline]
   fn next_back(&mut self) -> Option<Self::Item> {
-    let slice = &mut self.data.as_slice_mut()[self.base..self.len];
+    let slice =
+      &mut self.data.as_slice_mut()[self.base as usize..self.tail as usize];
     let item = slice.last_mut()?;
-    self.len -= 1;
+    self.tail -= 1;
     return Some(take(item));
   }
   #[cfg(feature = "rustc_1_40")]
   #[inline]
   fn nth_back(&mut self, n: usize) -> Option<Self::Item> {
-    let slice = &mut self.data.as_slice_mut()[self.base..self.len];
-    let n = slice.len().checked_sub(n + 1)?;
-    let item = &mut slice[n];
-    self.len = n;
+    let base = self.base as usize;
+    let tail = self.tail as usize;
+    let slice = &mut self.data.as_slice_mut()[base..tail];
+    let n = n.saturating_add(1);
 
-    return Some(take(item));
+    if let Some(n) = slice.len().checked_sub(n) {
+      let item = &mut slice[n];
+      /* n is in [0..self.tail - self.base] range, so in u16 range */
+      self.tail = self.base + n as u16;
+      return Some(take(item));
+    }
+
+    self.tail = self.base;
+    return None;
   }
 }
 
@@ -1171,7 +1247,7 @@ impl<A: Array> IntoIterator for ArrayVec<A> {
   #[inline(always)]
   #[must_use]
   fn into_iter(self) -> Self::IntoIter {
-    ArrayVecIterator { base: 0, len: self.len, data: self.data }
+    ArrayVecIterator { base: 0, tail: self.len, data: self.data }
   }
 }
 
@@ -1467,3 +1543,39 @@ impl<A: Array> ArrayVec<A> {
     self.drain_to_vec_and_reserve(0)
   }
 }
+
+#[cfg(feature = "serde")]
+struct ArrayVecVisitor<A: Array>(PhantomData<A>);
+
+#[cfg(feature = "serde")]
+impl<'de, A: Array> Visitor<'de> for ArrayVecVisitor<A>
+where
+  A::Item: Deserialize<'de>,
+{
+  type Value = ArrayVec<A>;
+
+  fn expecting(
+    &self,
+    formatter: &mut core::fmt::Formatter,
+  ) -> core::fmt::Result {
+    formatter.write_str("a sequence")
+  }
+
+  fn visit_seq<S>(self, mut seq: S) -> Result<Self::Value, S::Error>
+  where
+    S: SeqAccess<'de>,
+  {
+    let mut new_arrayvec: ArrayVec<A> = Default::default();
+
+    let mut idx = 0usize;
+    while let Some(value) = seq.next_element()? {
+      if new_arrayvec.len() >= new_arrayvec.capacity() {
+        return Err(DeserializeError::invalid_length(idx, &self));
+      }
+      new_arrayvec.push(value);
+      idx = idx + 1;
+    }
+
+    Ok(new_arrayvec)
+  }
+}
